function [ outputMatrix ] = getMatrix( n, shape, p, a, b )
%getMatrix Returns a nxn matrix in the .mtx-format that is upper triangular
%(shape='u'), lower triangular (shape='l') or diagonal (shape='d').
%The matrix has nonzero values on the diagonal and p percentage * n many
%values at random positions. The values are in the range [a,b].
%   n: Number of rows and columns of the quadratic matrix
%   shape: Shape of the matrix (upper triangular ('u'), lower triangular
%   ('l') or diagonal ('d').
%   p: The matrix has p percentage * n^2 many nonzero values at random
%   positions. p is a value between 0 and 100.
%   a: Lower limit of the interval in which the values are.
%   b: Upper limit of the interval in which the values are.
% Example usage: spy(getMatrix(100, 'u', 1 ,-100,100));

tic
if n<=0
    fprintf('Error in getMatrix-function: n shall be a positive integer!\n')
    return;
end

if rem(n,1)~=0
     fprintf('Error in getMatrix-function: n shall be a integer number!\n')
    return;
end

if shape=='u' || shape=='l' || shape=='d'
    %shape is valid and we can go on.
else
     fprintf('Error in getMatrix-function: shape should be ''u'', ''l'' or ''d''!\n')
    return;
end

if p<0 || p>100
     fprintf('Error in getMatrix-function: p is not between 0 and 100!\n')
    return;
end

if ~isnumeric(a) || ~isnumeric(b)
     fprintf('Error in getMatrix-function: a or b is not numeric!\n')
    return;
end

if shape=='d'
    
    randomNumbers = a + (b-a).*rand(n,1);

    l=randomNumbers(:)==0;
    l=find(l);                      %Array of indeces of zeros in d

    for i=1:length(l)
        randomNumbers(l(i))=a+(b-a)*rand;       %Fill zero entries of d with random values
    end

    % create a sparse diagonal matrix with place for n+p/100*n*n entries
    %outputMatrix=spdiags(d(:),0,length(d),length(d));
    outputMatrix=sparse(1:n,1:n,randomNumbers,n,n,n+round(p/100*n*n));
    %outputMatrix=diag(d);
    
    mmwrite('generatedMatrix.mtx',outputMatrix);
    return;     %We already have a diagonal matrix, so we stopp here.
end

noOfEntriesToAddNotOnDiagonal=round(p/100*n*n);
row_indices=[1:n zeros(1,noOfEntriesToAddNotOnDiagonal)];
column_indices=[1:n zeros(1,noOfEntriesToAddNotOnDiagonal)];
randomNumbers = a + (b-a).*rand(1,n+noOfEntriesToAddNotOnDiagonal);

%This is the upper triangular case
if shape=='u'
    k=1;
    while k<=noOfEntriesToAddNotOnDiagonal
        row=round(1+(n-1)*rand);
        column=round(1+(n-1)*rand);
        if column==row && column==n
                                            %Do nothing and start again
        elseif column==row && column~=n
            column=column+1;
            row_indices(n+k)=row;
            column_indices(n+k)=column;
            k=k+1;
        elseif column < row
            [column,row]=deal(row,column);          %Swap the values
            row_indices(n+k)=row;
            column_indices(n+k)=column;
            k=k+1;
        else
            row_indices(n+k)=row;
            column_indices(n+k)=column;
            k=k+1;
        end
    end
end

%This is the lower triangular case
if shape=='l'
    k=1;
    while k<=noOfEntriesToAddNotOnDiagonal
        row=round(1+(n-1)*rand);
        column=round(1+(n-1)*rand);
        if column==row && column==n
                                            %Do nothing and start again
        elseif column==row && column~=n
            row=row+1;
            row_indices(n+k)=row;
            column_indices(n+k)=column;
            k=k+1;
        elseif column > row
            [column,row]=deal(row,column);          %Swap the values
            row_indices(n+k)=row;
            column_indices(n+k)=column;
            k=k+1;
        else
            row_indices(n+k)=row;
            column_indices(n+k)=column;
            k=k+1;
        end
    end
end


outputMatrix = sparse(row_indices,column_indices,randomNumbers,n,n);
mmwrite('generatedMatrix.mtx',outputMatrix);

toc
end

