/*
    ======================================================================
    ======================================================================
    ==                                                                  ==
    ==  MSPAI:  Modified SPAI algorithm to comupte SParse Approximate   ==
    ==          Invers matrices.                                        ==
    ==                                                                  ==
    ==  Copyright (C)  2007, 2008, 2009 by                              ==
    ==                 Matous Sedlacek <sedlacek@in.tum.de>             ==
    ==                 Chair of Scientific Computing -- Informatics V   ==
    ==                 Technische Universität München                   ==
    ==                                                                  ==
    ==  This file is part of MSPAI.                                     ==
    ==                                                                  ==
    ==  MSPAI is free software: you can redistribute it and/or          ==
    ==  modify it under the terms of the GNU Lesser General Public      ==
    ==  License as published by the Free Software Foundation, either    ==
    ==  version 3 of the License, or (at your option) any later version.==
    ==                                                                  ==
    ==  MSPAI is distributed in the hope that it will be useful,        ==
    ==  but WITHOUT ANY WARRANTY; without even the implied warranty of  ==
    ==  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   ==
    ==  GNU Lesser General Public License for more details.             ==
    ==                                                                  ==
    ==  You should have received a copy of the GNU Lesser General       ==
    ==  Public License along with MSPAI.                                ==
    ==  If not, see <http://www.gnu.org/licenses/>.                     ==
    ==                                                                  ==
    ======================================================================
    ======================================================================
*/


#ifndef GUARD_MATRIX_H
#define GUARD_MATRIX_H



#include "Compressed_Lines.h"
#include "Matrix_Base.h"
#include "Pattern.h"


//////////////////////////////////////////
///     \brief Simple structure representing
///            a complex number   
//////////////////////////////////////////
struct COMPLEX
{
    /// real part of complex number
    double real;
    
    /// imaginary part of complex number
    double imag;    
};


/// Operator overloading for "*" between 
/// complex numbers
inline const COMPLEX 
operator*(const COMPLEX& a, const COMPLEX& b)
{
    COMPLEX out;
    out.real = a.real * b.real - a.imag * b.imag;
    out.imag = a.real * b.imag + a.imag * b.real;
    return out;
} 


/// Operator overloading for "+" between 
/// complex numbers
inline const COMPLEX 
operator+(const COMPLEX& a, const COMPLEX& b)
{
    COMPLEX out;
    out.real = a.real + b.real;
    out.imag = a.imag + b.imag;
    return out;
} 


/// Operator overloading for "-" between 
/// complex numbers
inline const COMPLEX 
operator-(const COMPLEX& a, const COMPLEX& b)
{
    COMPLEX out;
    out.real = a.real - b.real;
    out.imag = a.imag - b.imag;
    return out;
}


/// Operator overloading for "/" between 
/// complex numbers
inline const COMPLEX 
operator/(const COMPLEX& a, const double& b)
{
    COMPLEX out;
    out.real = a.real / b;
    out.imag = a.imag / b;
    return out;
} 


//////////////////////////////////////////
///     \class Matrix
///     \brief The Matrix datastructure
//////////////////////////////////////////
template <class T>
class Matrix : public Matrix_Base
{
  
    public:
        
        /// Empty Constructor
        Matrix<T>();
        
        /// Destructor
        ~Matrix<T>();
        
        
        //Member variables
        
        /// Compressed column storage (CCS)
        /// This datastructure contains the 
        /// matrix relevant data on this pe
        Compressed_Lines<T> *c_lines;                 
        
        
        //Methods
        
        //============================================================================
        //================ Template specifications for double matrices ===============
        //============================================================================
        
        /////////////////////////////////////////////////////////
        ///     \brief  Printing all matrix data 
        ///             
        ///     This method prints all data in shortest
        ///     screen output.
        ///
        ///     \param  matrix The matrix to be printed
        /////////////////////////////////////////////////////////
        void        Print_Matrix_Data(Matrix<double> *matrix);
        

        /////////////////////////////////////////////////////////
        ///     \brief  Printing the real matrix values
        ///             in matrix market format into a 
        ///             file.
        ///             
        ///     This method prints only the matrix values
        ///     into a file in matrix market format. 
        ///     This format is a simple row-column-value
        ///     format (e.g.:) 2 1 9.003
        ///
        ///     \param  matrix The matrix to be printed
        ///                    to file
        ///     \param file The file where the matrix
        ///                 should be printed to
        /////////////////////////////////////////////////////////
        void        Write_Matrix_To_File(   Matrix<double>  *matrix,
                                            char            *file);
        

        /////////////////////////////////////////////////////////
        ///     \brief  Printing the matrix values in 
        ///             human readable format
        ///             
        ///     This method prints only the matrix values
        ///     in human readable format.
        ///
        ///     \param  matrix The matrix to be printed
        ///     \param n n-dimension of the matrix (columns)
        ///     \param m m-dimension of the matrix (rows)
        /////////////////////////////////////////////////////////
        void        Print_Matrix_Human_Readable(const Matrix<double>    *matrix, 
                                                const int               n, 
                                                const int               m);
        
        
        /////////////////////////////////////////////////////////
		///     \brief  Printing the matrix values in a format
		///             that can be copied into Matlab
		///
		///     \param  matrix The matrix to be printed
		///     \param n n-dimension of the matrix (columns)
		///     \param m m-dimension of the matrix (rows)
		/////////////////////////////////////////////////////////
		void        Print_Matrix_For_Matlab    (const Matrix<double>    *matrix,
												const int               n,
												const int               m);



        //============================================================================
        //================ Template specifications for COMPLEX matrices ==============
        //============================================================================
        
        /////////////////////////////////////////////////////////
        ///     \brief  Printing all matrix data 
        ///             
        ///     This method prints all data in shortest
        ///     screen output.
        ///
        ///     \param  matrix The matrix to be printed
        /////////////////////////////////////////////////////////
        void        Print_Matrix_Data(Matrix<COMPLEX> *matrix);
        
        
        /////////////////////////////////////////////////////////
        ///     \brief  Printing the matrix values in 
        ///             human readable format
        ///             
        ///     This method prints only the matrix values
        ///     in human readable format.
        ///
        ///     \param  matrix The matrix to be printed
        ///     \param n n-dimension of the matrix (columns)
        ///     \param m m-dimension of the matrix (rows)
        /////////////////////////////////////////////////////////
        void        Print_Matrix_Human_Readable(const Matrix<COMPLEX>   *matrix, 
                                                const int               n, 
                                                const int               m);
        
        
        /////////////////////////////////////////////////////////
        ///     \brief  Printing the complex matrix 
        ///             values in matrix market format 
        ///             into a file.
        ///             
        ///     This method prints only the matrix values
        ///     into a file in matrix market format. 
        ///     This format is a simple row-column-value
        ///     format (e.g.:) 2 1 9.003 9.884
        ///
        ///     \param  matrix The matrix to be printed
        ///                    to file
        ///     \param file The file where the matrix
        ///                 should be printed to
        /////////////////////////////////////////////////////////
        void        Write_Matrix_To_File(   Matrix<COMPLEX>     *matrix,
                                            char                *file);

        
        
        //============================================================================
        //===================== Template methods - see Matrix.imp ====================
        //============================================================================
        
        /////////////////////////////////////////////////////////
        ///     \brief  Initializing preconditioner with
        ///             default values
        ///
        ///     \param  M The preconditioner to be 
        ///               intialized
        ///     \param m_dim Number of rows M should have
        ///     \param n_dim Number of columns M should have
        /////////////////////////////////////////////////////////
        void            Init_Preconditioner(Matrix<T> *&M, 
                                            const int m_dim, 
                                            const int n_dim);


        /////////////////////////////////////////////////////////
        ///     \brief  Generating Pattern out of Matrix
        ///
        ///     \param  mtx The matrix from which the pattern 
        ///                 will be generated from
        ///     \param use_prob Whether probing is requested or 
        ///                     not.
        ///     \return Generated Pattern from Matrix
        /////////////////////////////////////////////////////////
        Pattern*        To_Pattern(Matrix<T> *mtx, 
                                   const bool use_prob);
        
        
        /////////////////////////////////////////////////////////
        ///     \brief  Generating tril(A) and tiru(A)
        ///
        ///     \param  mtx The matrix from which tril and triu
        ///                 will be generated from
        ///     \param L will be tril(A)
        ///     \param U will be triu(A)
        /////////////////////////////////////////////////////////
        void        trilu(Matrix<T> *A, Matrix<T> *&L, Matrix<T> *&U);

    private:
        
        /////////////////////////////////////////////////////////
        ///     \brief  Counting the nnz elements of 
        ///             this matrix
        ///
        ///     \return Number of nnz within this matrix
        /////////////////////////////////////////////////////////
        int             Count_NNZ();
                
};

#include "Matrix.imp"

#endif
