/*
    ======================================================================
    ======================================================================
    ==                                                                  ==
    ==  MSPAI:  Modified SPAI algorithm to comupte SParse Approximate   ==
    ==          Invers matrices.                                        ==
    ==                                                                  ==
    ==  Copyright (C)  2007, 2008, 2009 by                              ==
    ==                 Matous Sedlacek <sedlacek@in.tum.de>             ==
    ==                 Chair of Scientific Computing -- Informatics V   ==
    ==                 Technische Universität München                   ==
    ==                                                                  ==
    ==  This file is part of MSPAI.                                     ==
    ==                                                                  ==
    ==  MSPAI is free software: you can redistribute it and/or          ==
    ==  modify it under the terms of the GNU Lesser General Public      ==
    ==  License as published by the Free Software Foundation, either    ==
    ==  version 3 of the License, or (at your option) any later version.==
    ==                                                                  ==
    ==  MSPAI is distributed in the hope that it will be useful,        ==
    ==  but WITHOUT ANY WARRANTY; without even the implied warranty of  ==
    ==  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   ==
    ==  GNU Lesser General Public License for more details.             ==
    ==                                                                  ==
    ==  You should have received a copy of the GNU Lesser General       ==
    ==  Public License along with MSPAI.                                ==
    ==  If not, see <http://www.gnu.org/licenses/>.                     ==
    ==                                                                  ==
    ======================================================================
    ======================================================================
*/


#ifndef GUARD_PATTERN_H
#define GUARD_PATTERN_H


//file includings
#include "Index_Set.h"


//C++ includings
#include <stdio.h>
#include <stdlib.h>


//////////////////////////////////////////
///     \brief Row Column structure 
///            representing a row column 
///            format for pattern
//////////////////////////////////////////
struct RC
{
    int i;
    int j;
};


//////////////////////////////////////////
///     \class Pattern
///     \brief The Pattern datastructure
//////////////////////////////////////////
class Pattern
{
  
    public:
        
        /// Empty Constructor
        Pattern() { };
        
        /// Constructor
        Pattern(int len);
        
        /// Destructor
        ~Pattern();
        
        
        //Member variables
        
        /// 2D array containing index 
        /// sets of each pattern column
        Index_Set   **j_sets;
        
        /// The number of columns/index sets
        /// this pattern contains
        int         len; 


        //Methods       
        
        //////////////////////////////////////////////////////
        ///     \brief  Printing all pattern data 
        //////////////////////////////////////////////////////
        void        Print_Pattern_Data();
        
        
        //////////////////////////////////////////////////////
        ///     \brief  Building pattern if user wants
        ///             unit matrix as input pattern
        ///
        ///     \param n The dimension of the pattern
        ///     \return The diagonal pattern
        //////////////////////////////////////////////////////
        Pattern*    Diagonal_Pattern(const int n);


        //////////////////////////////////////////////////////
        ///     \brief  Building pattern out of pattern file
        ///
        ///     Building pattern datastructure out of pattern
        ///     file. If user wants P = patt(A) (option -2), 
        ///     then parse the matrix file and discard the 
        ///     the third matrix value.
        ///
        ///     \param file Path to file to build pattern from
        ///     \param matrix_dim Dimension of the input matrix
        ///     \param use_schur Whether to use schur probing
        ///                      or not
        ///     \param prob_Ce_N Number of columns the Ce
        ///                      probing matrix have
        ///     \return The generated pattern.
        //////////////////////////////////////////////////////
        Pattern*    Arbitrary_Pattern(  char        *file,
                                        int         matrix_dim,
                                        const int   use_schur,
                                        const int   prob_Ce_N);
        
    private:        
        
        //////////////////////////////////////////////////////
        ///     \brief  Open pattern file and reading pattern
        ///             header.
        ///
        ///     \param f Path to the pattern file matrix
        ///     \param file_M m-dimension of the pattern matrix
        ///     \param file_N n-dimension of the pattern matrix
        ///     \param file_nnz Number of nnzs within this 
        ///                     pattern
        ///     \param symmetric Whether the pattern is symmetric
        ///                      or not.
        //////////////////////////////////////////////////////
        void        Read_Pattern_Header(FILE *&f, 
                                        int &file_M, 
                                        int &file_N, 
                                        int &file_nnz,
                                        bool &symmetric);
        
        
        //////////////////////////////////////////////////////
        ///     \brief  Parsing file and filling arrays with
        ///             row/column elements specific to this
        ///             pe.
        ///
        ///     \param f Path to the pattern file matrix
        ///     \param file_nnz Number of nnz within the pattern
        ///                     file
        ///     \param my_nbr_cols Number of columns of this pe
        ///     \param start_idx Start index of this pe's 
        ///                      local work chunk within whole
        ///                      pattern
        ///     \param cols The row/column structure elements
        ///                 read out from pattern file
        ///     \param nnz_cols Number of nnzs in each pattern
        ///                     column
        ///     \param symmetric Whether the pattern is symmetric
        ///                      or not.
        //////////////////////////////////////////////////////
        void        Read_Pattern_Data(  FILE *f,
                                        int file_nnz,
                                        int my_nbr_cols,
                                        int start_idx,
                                        RC  **cols,
                                        int *nnz_cols,
                                        const bool symmetric);
        
        
        //////////////////////////////////////////////////////
        ///     \brief  Skipping Header to get relevant
        ///             pattern data
        //////////////////////////////////////////////////////
        void        Skip_Header(FILE *f);
        
        
        //////////////////////////////////////////////////////
        ///     \brief  Mapping read out pattern data to 
        ///             Pattern data structure
        ///
        ///     \param file_N Dimension of the pattern matrix
        ///     \param cols The row/column structure elements
        ///                 read out from pattern file
        ///     \param nnz_cols Number of nnzs in each pattern
        ///                     column
        ///     \return The built pattern
        //////////////////////////////////////////////////////
        Pattern*    Data_To_Pattern(int file_N,
                                    RC  *cols,
                                    int nnz_cols);  
        
        
        //////////////////////////////////////////////////////
        ///     \brief  Counting nnz in each column 
        ///
        ///     \param in array containing the RC elements
        ///               read out from the pattern file
        ///     \param size size of the in array
        ///     \param out array containing the number of
        ///                nnzs of each column within the 
        ///                pattern
        //////////////////////////////////////////////////////
        void        Count_NNZ_Cols( RC const *in, 
                                    size_t size, 
                                    int *out);
        
        
        //////////////////////////////////////////////////////
        ///     \brief  Counting nnz in each row
        ///
        ///     \param in array containing the RC elements
        ///               read out from the pattern file
        ///     \param size size of the in array
        ///     \param out array containing the number of
        ///                nnzs of each row within the 
        ///                pattern
        //////////////////////////////////////////////////////
        void        Count_NNZ_Rows( RC const *in, 
                                    size_t size, 
                                    int *out);
};

#endif
